﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Net.Http;
using System.Text;
using ELTE.TodoList.Desktop.Model;
using ELTE.TodoList.Persistence.DTO;

namespace ELTE.TodoList.Desktop.ViewModel
{
    public class MainViewModel : ViewModelBase
    {
        private readonly TodoListApiService _service;
        private ObservableCollection<ListViewModel> _lists;
        private ObservableCollection<ItemViewModel> _items;
        private ItemViewModel _selectedItem;
        private ListViewModel _selectedList;
        private ItemViewModel _editableItem;
        private String _selectedListName;

        public ObservableCollection<ListViewModel> Lists
        {
            get { return _lists; }
            set { _lists = value; OnPropertyChanged(); }
        }

        public ObservableCollection<ItemViewModel> Items
        {
            get { return _items; }
            set { _items = value; OnPropertyChanged(); }
        }

        public ItemViewModel SelectedItem
        {
            get { return _selectedItem; }
            set { _selectedItem = value; OnPropertyChanged(); }
        }

        public ListViewModel SelectedList
        {
            get { return _selectedList; }
            set { _selectedList = value; OnPropertyChanged(); }
        }

        public String SelectedListName
        {
            get { return _selectedListName; }
            set { _selectedListName = value; OnPropertyChanged(); }
        }

        public ItemViewModel EditableItem
        {
            get { return _editableItem; }
            set { _editableItem = value; OnPropertyChanged(); }
        }

        public DelegateCommand RefreshListsCommand { get; private set; }
        public DelegateCommand SelectCommand { get; private set; }
        public DelegateCommand LogoutCommand { get; private set; }
        public DelegateCommand AddListCommand { get; private set; }
        public DelegateCommand EditListCommand { get; private set; }
        public DelegateCommand DeleteListCommand { get; private set; }
        public DelegateCommand AddItemCommand { get; private set; }
        public DelegateCommand EditItemCommand { get; private set; }
        public DelegateCommand DeleteItemCommand { get; private set; }
        public DelegateCommand SaveItemEditCommand { get; private set; }
        public DelegateCommand CancelItemEditCommand { get; private set; }
        public DelegateCommand ChangeImageCommand { get; private set; }

        #region Events

        public event EventHandler LogoutSucceeded;

        public event EventHandler StartingItemEdit;

        public event EventHandler FinishingItemEdit;

        public event EventHandler StartingImageChange;

        #endregion

        public MainViewModel(TodoListApiService service)
        {
            _service = service;

            RefreshListsCommand = new DelegateCommand(_ => LoadListsAsync());
            SelectCommand = new DelegateCommand(_ => LoadItemsAsync(SelectedList));
            LogoutCommand = new DelegateCommand(_ => LogoutAsync());

            AddListCommand = new DelegateCommand(_ => AddList());
            EditListCommand = new DelegateCommand(_ => !(SelectedList is null), _ => EditList());
            DeleteListCommand = new DelegateCommand(_ => !(SelectedList is null), _ => DeleteList(SelectedList));

            AddItemCommand = new DelegateCommand(_ => !(SelectedList is null), _ => AddItem());
            EditItemCommand = new DelegateCommand(_ => !(SelectedItem is null), _ => StartEditItem());
            DeleteItemCommand = new DelegateCommand(_ => !(SelectedItem is null), _ => DeleteItem(SelectedItem));

            SaveItemEditCommand = new DelegateCommand(_ => SaveItemEdit());
            CancelItemEditCommand = new DelegateCommand(_ => CancelItemEdit());
            ChangeImageCommand = new DelegateCommand(_ => StartingImageChange?.Invoke(this, EventArgs.Empty));
        }

        private async void LogoutAsync()
        {
            try
            {
                await _service.LogoutAsync();
                LogoutSucceeded?.Invoke(this, EventArgs.Empty);
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }

        }

        #region Lists

        private async void LoadListsAsync()
        {
            try
            {
                Lists = new ObservableCollection<ListViewModel>((await _service.LoadListsAsync())
                    .Select(list => (ListViewModel)list));
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private async void AddList()
        {
            var newList = new ListViewModel
            {
                Name = SelectedListName
            };

            var listDto = (ListDto)newList;

            try
            {
                await _service.CreateListAsync(listDto);
                newList.Id = listDto.Id;
                Lists.Add(newList);
                SelectedList = newList;
            }

            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private async void DeleteList(ListViewModel list)
        {
            try
            {
                await _service.DeleteListAsync(list.Id);
                Lists.Remove(SelectedList);
                SelectedList = null;
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private async void EditList()
        {
            try
            {
                SelectedList.Name = SelectedListName;
                await _service.UpdateListAsync((ListDto)SelectedList);
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }

        #endregion

        #region Items

        private async void LoadItemsAsync(ListViewModel list)
        {
            if (list is null)
            {
                Items = null;
                return;
            }

            try
            {
                SelectedListName = list.Name;
                Items = new ObservableCollection<ItemViewModel>((await _service.LoadItemsAsync(list.Id))
                    .Select(item => (ItemViewModel)item));

            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private async void AddItem()
        {
            var newItem = new ItemViewModel
            {
                Name = "New Item",
                Deadline = DateTime.Now,
                ListId = SelectedList.Id
            };

            var itemDto = (ItemDto)newItem;

            try
            {
                await _service.CreateItemAsync(itemDto);
                newItem.Id = itemDto.Id;
                Items.Add(newItem);
                SelectedItem = newItem;
            }

            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private async void DeleteItem(ItemViewModel item)
        {
            try
            {
                await _service.DeleteItemAsync(item.Id);
                Items.Remove(SelectedItem);
                SelectedItem = null;
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
        }
        private void StartEditItem()
        {
            EditableItem = SelectedItem.ShallowClone();
            StartingItemEdit?.Invoke(this, EventArgs.Empty);
        }
        private void CancelItemEdit()
        {
            EditableItem = null;
            FinishingItemEdit?.Invoke(this, EventArgs.Empty);
        }
        private async void SaveItemEdit()
        {
            try
            {
                SelectedItem.CopyFrom(EditableItem);
                await _service.UpdateItemAsync((ItemDto)SelectedItem);
                if (SelectedItem.ListId != SelectedList.Id)
                {
                    Items.Remove(SelectedItem);
                    SelectedItem = null;
                }
            }
            catch (Exception ex) when (ex is NetworkException || ex is HttpRequestException)
            {
                OnMessageApplication($"Unexpected error occurred! ({ex.Message})");
            }
            FinishingItemEdit?.Invoke(this, EventArgs.Empty);
        }

        #endregion
    }
}
